# Section 14.2 Example 1

def sum_vectors(_u, _v):
    """Return the sum of two n-dimensional vectors _u and _v."""
    return [_u[i] + _v[i] for i in range(len(_u))]

def k_times_vector(_u, k):
    """Return the n-dimensional vector _u multiplied by k."""
    return [k*u for u in _u]

# Section 14.2 Example 2

def dot_product(_u, _v):
    return sum(_u[i]*_v[i] for i in range(len(_u)))

# Section 14.3 Example 2

def empty_matrix(n_rows, n_cols, x=0):
    """Return an n_rows by n_cols matrix with all elements set to x."""
    m = []
    for r in range(n_rows):
        m.append(n_ncols*[x])
    return m

# Section 14.3 Question 1

def trace(a):
    """ Return the trace of the matrix a."""
    return sum(a[i][i] for i in range(len(a)))

if __name__ == '__main__':  # Run only as a stand-alone program,
                            # not when imported into other programs
    a = [[1.0, 2.0, 3.0],
         [0.4, 0.5, 0.6],
         [0.07, 0.08, 0.09]]

    print(trace(a))

# Section 14.3 Question 2

def display_matrix(m):
    """Print out m with aligned columns."""
    for row in m:
        for x in row:
            print('{0:4d}'.format(x), end = '')
        print()
    print()

if __name__ == '__main__':
    m = [[1, 2, 3],
         [11, 21, 31],
         [21, 22, 33]]
    display_matrix(m)

# Section 14.3 Question 3

def read_matrix(pathname):
    """Read a matrix from a text file and return it."""
    f = open(pathname)
    m = []
    for line in f:
        row = [int(x) for x in line.replace('\t', ' ').replace(',', ' ').split()]
        m.append(row)
    f.close()
    return m

if __name__ == '__main__':
    m = read_matrix('matrix.txt')
    display_matrix(m)

# Section 14.3 Question 4

def transpose(a):
    """Return transposed a."""
    n = len(a)
    t = []
    for r in range(n):
        t.append(n*[0])
    for r in range(n):
        for c in range(n):
            t[c][r] = a[r][c]
    return t

if __name__ == '__main__':
    a = [[1, 2, 3],
         [11, 21, 31],
         [21, 22, 33]]

    display_matrix(a)
    display_matrix(transpose(a))

# Section 14.3 Question 5

def matrix_dot_vector(A, u):
    """Return the vector Av."""
    n = len(u)
    return [sum(A[r][c]*u[c] for c in range(n)) for r in range(n)]

# Or:

##def matrix_dot_vector(A, u):
##    """Return the vector Av."""
##    return [dot_product(row, u) for row in A]

if __name__ == '__main__':
    A = [[1, 2, 3],
         [4, 5, 6],
         [7, 8, 9]]

    u = (100, 10, 1)
    v = matrix_dot_vector(A, u)
    print(v)

# Section 14.3 Question 6

def matrix_dot_matrix(a, b):
    """Return a*b."""
    n = len(a)
    product = []
    for r in range(n):
        product.append(n*[0])
    for r in range(n):
        for c in range(n):
            product[r][c] = sum(a[r][i]*b[i][c] for i in range(n))
    return product

if __name__ == '__main__':
    A = [[1, 2, 3],
         [4, 5, 6],
         [7, 8, 9]]

    B = [[1, 2, 3],
         [2, 3, 1],
         [3, 1, 2]]

    display_matrix(matrix_dot_matrix(A, B))

# Section 14.3 Question 10

def determinant(a):
    """Return the determinant of the 2 by 2 matrix a."""
    return a[0][0]*a[1][1] - a[1][0]*a[0][1]

def solve(a, c):
    """ Solve the system of two linear equations ax = c."""
    d = determinant(a)
    if d == 0:
        return None, None
    x = determinant([[c[0], a[0][1]], [c[1], a[1][1]]]) / d
    y = determinant([[a[0][0], c[0]], [a[1][0], c[1]]]) / d
    return [x, y]

if __name__ == '__main__':
    a = [[2, -7],
         [3, -4]]

    print('determinant =', determinant(a))
    x, y = solve(a, [-11, 3])
    print('x =', x, 'y =', y)

# Section 14.3 Question 11

def determinant3(a):
    return (a[0][0]*a[1][1]*a[2][2] +
            a[0][1]*a[1][2]*a[2][0] +
            a[1][0]*a[2][1]*a[0][2] -
            a[0][2]*a[1][1]*a[2][0] -
            a[0][1]*a[1][0]*a[2][2] -
            a[2][1]*a[1][2]*a[0][0])

def copy_matrix(a):
    """Return a copy of a."""
    return [a[0][:], a[1][:], a[2][:]]

def solve3(a, c):
    """ Solve the system of three linear equations ax = c."""
    solution = 3*[None]
    d = determinant3(a)
    if d == 0:
        return solution
    for i in range(3):
        b = copy_matrix(a)
        b[0][i] = c[0]
        b[1][i] = c[1]
        b[2][i] = c[2]
        solution[i] = determinant3(b) / d
    return solution

if __name__ == '__main__':
    a = [[2, 3, 1],
         [4, 6, 5],
         [7, 9, 8]]

    c = [42, 96, 150]
    print('determinant = ', determinant3(a))
    print(solve3(a, c))

# Section 17.6 Question 8

def pow_matrix(A, n):
    """Return A^n."""
    Y = [[1, 0],
         [0, 1]]
    while n != 0:
        if n & 1 != 0:
            Y = matrix_dot_matrix(Y, A)
        A = matrix_dot_matrix(A, A)
        n >>= 1
    return Y
